"use strict";
exports.__esModule = true;
exports.answer = exports.initialSRSData = void 0;
var utils_1 = require("./utils");
var score_1 = require("./score");
var EASE_DELTAS = {
    WRONG: -0.2,
    HARD: -0.15,
    GOOD: 0,
    EASY: 0.15
};
exports.initialSRSData = function () { return ({
    lastAnsweredAt: null,
    dueAt: null,
    ease: 2.5,
    intervalInMins: null,
    reps: 0,
    lapses: 0,
    repsSinceLastLapse: 0
}); };
var DEFAULT_FUZZ_PERCENT = 0.05;
var DEFAULT_FUZZ_PROVIDER = function (i) {
    var fuzzinessFactor = 1 + Math.random() * 2 * DEFAULT_FUZZ_PERCENT - DEFAULT_FUZZ_PERCENT;
    return i * fuzzinessFactor;
};
/**
 * Calculate updated srs data.
 *
 * @param srsData The current srs data. This object will not be modified.
 * @param score The score (0, 1, 2 or 3)
 * @param answeredAt When the item was ansewered. This might differ from the time the updated srs data is calculated and is needed to calculate readiness.
 * @param fuzzProvider For testing only
 */
exports.answer = function (srsData, score, answeredAt, fuzzProvider) {
    if (fuzzProvider === void 0) { fuzzProvider = DEFAULT_FUZZ_PROVIDER; }
    if (!score_1.ALL_SCORES.includes(score)) {
        throw new TypeError("Invalid score: " + score);
    }
    var lastAnsweredAt = srsData.lastAnsweredAt, dueAt = srsData.dueAt, ease = srsData.ease, intervalInMins = srsData.intervalInMins, reps = srsData.reps, lapses = srsData.lapses, repsSinceLastLapse = srsData.repsSinceLastLapse;
    // The readiness factor varies depending on how far before or after its due date this item was answered.
    var readinessFactor = 1;
    if (dueAt && lastAnsweredAt) {
        var scheduledInterval = +dueAt - +lastAnsweredAt;
        var actualInterval = +answeredAt - +lastAnsweredAt;
        readinessFactor = actualInterval / scheduledInterval;
    }
    if (readinessFactor < 0) {
        throw new TypeError("readinessFactor must not be negative; srsData: " + JSON.stringify(srsData) + "; score: " + score);
    }
    var easeDelta = EASE_DELTAS[score] * readinessFactor;
    var updatedEase = utils_1.clamp(ease + easeDelta, 1.3, 3.5);
    // Update interval
    var updatedInterval;
    if (reps === 0 || score === 'WRONG') {
        updatedInterval = utils_1.daysInMins(1);
    }
    else if (repsSinceLastLapse === 1) {
        updatedInterval = utils_1.daysInMins(6);
    }
    else {
        if (intervalInMins === null) {
            throw new TypeError("intervalInMins is null when it shouldn't be");
        }
        var intervalDelta = intervalInMins * (updatedEase - 1);
        intervalDelta *= readinessFactor;
        intervalDelta = fuzzProvider(intervalDelta);
        updatedInterval = intervalInMins + intervalDelta;
        updatedInterval = Math.ceil(updatedInterval);
    }
    // Update rep counts
    var updatedReps = reps + 1;
    var updatedLapses, updatedRepsSinceLastLapse;
    if (score === 'WRONG') {
        updatedLapses = lapses + 1;
        updatedRepsSinceLastLapse = 0;
    }
    else {
        updatedLapses = lapses;
        updatedRepsSinceLastLapse = repsSinceLastLapse + 1;
    }
    // Calculate the next due at date
    var updatedDueAt = new Date(answeredAt.getTime() + updatedInterval * 60 * 1000);
    return {
        lastAnsweredAt: answeredAt,
        dueAt: updatedDueAt,
        ease: updatedEase,
        intervalInMins: updatedInterval,
        reps: updatedReps,
        lapses: updatedLapses,
        repsSinceLastLapse: updatedRepsSinceLastLapse
    };
};
